#####################################################################
# 
# Dispatcher.py - Polling Thread
#
# $Id: Dispatcher.py,v 1.3 2003/11/22 14:20:41 kevin Exp $
#
# Copyright (C) 2003 Kevin Gill
#
# Derived from the Xron version 0.0.10
# Copyright (C) 2000 Loren Stafford
#  
# Xron is Derived from the ZScheduler product version 0.0.7 in
# accord with the terms of the ZScheduler's license.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License  
# along with this program; if not, write to the 
#     Free Software Foundation, Inc., 
#     59 Temple Place - Suite 330, 
#     Boston, MA  02111-1307, USA.
# 
#####################################################################

""" Implementation of the Polling Thread

Works with the TQPollingThread object which keeps a list of 
objects to be polled. 

This thread loops around and calls the Process() method
of each object every 10 seconds (configurable).

"""

import Globals
from threading import Thread, Event
import ThreadedAsync
import sys

_python_version=float("%d.%d" % (sys.version_info[0], sys.version_info[1]))

class Dispatcher(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        Thread.__init__(self,group=group,target=target,name=name,args=args,kwargs=kwargs)
        self.setDaemon(1) # so thread will terminate with parent process
        if _python_version > 2.1:
            self.start()
        else:
            ## Assume that this is before Zope 2.7 if python 2.1 is used
            ThreadedAsync.register_loop_callback(self.becomeAsync)

    def becomeAsync(self, map):
        self.start()

def Timer(GlobalID):
    try:
        # Some imports are in this function. Otherwise a reload overwrites them.
        # Which causes the loop to fail.
        import zLOG, string

        # Use this for timing
        _ThreadEvent = Event()

        if _python_version > 2.1:
            # Give the core time to start
            Event.wait(10)

        maxwait=float(10)  # max time to wait between wake-ups in seconds
        minwait=float(0.1)  # max time to wait between wake-ups in seconds
        interval=maxwait # Initial sleep time. 

        # Private data is used to track data for each scheduled object
        PrivateData = {}

        # This is where we get our database connection - 
        DB = Globals.DB

        # infinite loop
        while 1:

            ### Connect to Zope, and locate the Polling Thread object
            conn = DB.open()
            root = conn.root()
            app  = root['Application']
            try: 
                PollingThread = getattr(app, GlobalID, None)
            except:
                zLOG.LOG('PollingThread', zLOG.PANIC, 'Cannot access /%s. Suspending operation.'% GlobalID)
                return

            ### Reload configurable items from the PollingThread
            try:
                maxwait = float(PollingThread.maxInterval)/1000
                minwait = float(PollingThread.minInterval)/1000
            except:
                pass

            ### Loop through the queues and find the items to process

            for objectPath in PollingThread:
                try:
                    # Create an instance
                    instance = app.unrestrictedTraverse(objectPath)
                except:
                    zLOG.LOG('PollingThread', zLOG.WARNING, "Polled Object %s not found - assuming deleted" % objectPath)
                    PollingThread.unregisterObject(objectPath)
                    get_transaction().commit()
                    continue

                # Ensure that the object has a Process Method
                try:
                    Method = instance.Process
                except:
                    zLOG.LOG('PollingThread', zLOG.WARNING, "Polled Object %s has no Process Method" % objectPath)
                    PollingThread.unregisterObject(objectPath)
                    get_transaction().commit()
                    continue

                # Retrieve the Private Data
                if PrivateData.has_key(objectPath):
                    PD = PrivateData[objectPath]
                else:
                    PD = {}
                    PrivateData[objectPath] = PD

                try:
                    rv = Method(App=app, private_data=PD)
                    if rv:
                        interval=minwait
                    get_transaction().commit()
                except:
                    import sys, traceback
                    (type, val, tb) = sys.exc_info()
                    zLOG.LOG('PollingThread', zLOG.ERROR, "Polled Object %s - Exception Occurred\n%s %s\n%s\n" %
                        (objectPath, type, val,
                        string.join(traceback.format_exception(type, val, tb), '\n')))
                    get_transaction().abort()
                    continue

                del instance

            # Close the Connection to Zope
            conn.close()
            del app

            #  Sleep for predetermined interval
            _ThreadEvent.wait(interval)
            interval=maxwait # Reset the default sleep time. May be recalculated above

    finally:
        # End of outer while 1 loop.
        # Something bad happened, let's clean up before quitting for good
        try:
            conn.close()
            del conn
            del app
        except:
            pass

