
## Stress Tester - See README.txt for details
## $Id: Driver.py,v 1.2 2003/11/23 23:37:56 kevin Exp $

import httplib, urllib
from random import random

# ----------------------------------------------------------------------------
# These are the configuration of the connection to the host

HOST='localhost'
HOST='127.0.0.1'  # Avoids  DNS Access
PORT='80'
PATH="/test/StressTest/"

pageDequeue = 'DequeueItems'
pageEnqueue = 'QueueItems'
pageQueueClear = 'QueueClear'
pageQueueGetSize = 'QueueGetSize'

# These are the configuration of the test

READTHREADS=10         # Number of threads reading
WRITETHREADS=10        # Number of threads writing
PERTHREADWRITE=5       # Number of writes per write thread (average)
PERTHREADREAD=5        # Number of reads per read thread (average)
REPEATS=5000           # Number of repeats per thread

# ----------------------------------------------------------------------------
# The following 4 functions are the basic building blocks for
# manipulating the queue.
#

def GetQueueSize():
    conn = httplib.HTTPConnection(HOST, PORT)
    try:
        conn.request("GET", PATH + pageQueueGetSize)
        r = conn.getresponse()
        assert r.status == 200
        return int(r.read())
    finally:
        conn.close()

def ClearQueue():
    conn = httplib.HTTPConnection(HOST, PORT)
    try:
        conn.request("GET", PATH + pageQueueClear)
        r = conn.getresponse()
        assert r.status == 200
        return 
    finally:
        conn.close()

def Enqueue(qnum, qmsg, qcnt):
    conn = httplib.HTTPConnection(HOST, PORT)
    try:
        params = urllib.urlencode({'qnum': qnum, 'qmsg': qmsg, 'qcnt': qcnt})
        headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
        conn.request("POST", PATH + pageEnqueue, params, headers)
        r = conn.getresponse()
        assert r.status == 200
        result = r.read().split('\n')
        rv = []
        for line in result:
            if len(line) == 0: continue
            rv.append(int(line))
        return rv
    finally:
        conn.close()

def Dequeue(qnum):
    conn = httplib.HTTPConnection(HOST, PORT)
    try:
        params = urllib.urlencode({'qnum': qnum})
        headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
        conn.request("POST", PATH + pageDequeue, params, headers)
        r = conn.getresponse()
        assert r.status == 200
        result = r.read().split('\n')
        rv = []
        for line in result:
            if len(line) == 0: continue
            if line == 'None': continue
            rv.append(eval(line))
        return rv
    finally:
        conn.close()

# ----------------------------------------------------------------------------
#   A single test thread will loop through a number of times.
#   It will either enqueue and dequeue a number items
#   It will keep track of the OIDs of the items enqueued/dequeued

import threading, time

ALLTHREADS=READTHREADS + WRITETHREADS

# Start is used to control when the test starts
# Finish recognises when the test ends
mutexStart = threading.BoundedSemaphore(ALLTHREADS)
mutexFinish = threading.BoundedSemaphore(ALLTHREADS)

class Tester(threading.Thread):

    def __init__(self, read_or_write, count, repeats, qmsg="Tester"):
        threading.Thread.__init__(self)
        self.read_or_write = read_or_write
        self.count = count
        self.repeats = repeats
        self.qcnt = 0
        self.qmsg = qmsg
        self.total = 0
        self.OIDS=[]

    def run(self):
        mutexFinish.acquire()
        sema = mutexStart.acquire()
        startTime = time.time()
        try:
            for i in range(self.repeats):
                count = int(random() * self.count * 2)
                if count == 0: count =1
                if self.read_or_write == 'w':
                    res = Enqueue(count, self.qmsg, self.qcnt)
                    self.qcnt += count
                    self.OIDS += res
                else:
                    res = Dequeue(count)
                    self.qcnt += len(res)
                    self.OIDS += map(lambda x:x[0], res)
        finally:
            mutexStart.release()
            mutexFinish.release()
            ## Log the finish of the threads
            print "Thread Finished (%s), (time=%.2f seconds) (%d)" % (self.read_or_write,
                time.time() - startTime, self.qcnt)

# ----------------------------------------------------------------------------
# And here is the control logic

# Clear the Queue
ClearQueue()

# Acquire all the start threads, so the threads cannot start
for i in range(ALLTHREADS):
    mutexStart.acquire()

# Set up the threads
write_threads=[]
read_threads=[]
if WRITETHREADS > READTHREADS:
    minStart = READTHREADS
else:
    minStart = WRITETHREADS

for i in range(minStart):
    t = Tester('w', PERTHREADWRITE, REPEATS)
    t.start()
    write_threads.append(t)
    t = Tester('r', PERTHREADREAD, REPEATS)
    t.start()
    read_threads.append(t)
for i in range(minStart, WRITETHREADS):
    t = Tester('w', PERTHREADWRITE, REPEATS)
    t.start()
    write_threads.append(t)
for i in range(minStart, READTHREADS):
    t = Tester('r', PERTHREADREAD, REPEATS)
    t.start()
    read_threads.append(t)

print "Starting...."
startTime = time.time()

# Release all the start threads
for i in range(ALLTHREADS):
    mutexStart.release()
# Wait for all the threads to finish
for i in range(ALLTHREADS):
    mutexFinish.acquire()

endTime = time.time()

# ----------------------------------------------------------------------------
# Test is finished - Look at the results

# Accumulate the results, and display
countRead = countWritten = 0
writeOIDS = []
readOIDS = []
for t in write_threads:
    countWritten += t.qcnt
    writeOIDS += t.OIDS
for t in read_threads:
    countRead += t.qcnt
    readOIDS += t.OIDS

print "Read", countRead
print "Written", countWritten
print "Net", countWritten - countRead
print "Queue Final", GetQueueSize()
print "Duration", endTime - startTime

if GetQueueSize() != countWritten - countRead:
    print "*** ERROR - RESULTS DO NOT MATCH ***"

readOIDS.sort()
writeOIDS.sort()
remaining = GetQueueSize()
if remaining == 0:
    remainingOIDS=[]
else:
    remainingOIDS = map(lambda x: x[0], Dequeue(remaining))
TotalRead = readOIDS + remainingOIDS
TotalRead.sort()

for k in readOIDS:
    if readOIDS.count(k) != 1:
        print "ERROR in Reads, OID %s Occurs %d times" % (k, readOIDS.count(k))
for k in writeOIDS:
    if writeOIDS.count(k) != 1:
        print "ERROR in Writes, OID %s Occurs %d times" % (k, writeOIDS.count(k))
for k in readOIDS:
    if writeOIDS.count(k) != 1:
        print "ERROR in Write/Reads, OID %s was written %d times" % (k, writeOIDS.count(k))

if readOIDS != writeOIDS[:len(readOIDS)]:
    print "Error, the items read back to not match the first %d items written" % len(readOIDS)
    for i in range(len(readOIDS)):
        if readOIDS[i] != writeOIDS[i]:
            print "Difference at position %d,\nread: %s\nwrote: %s" % (i,
                readOIDS[i:i+5], writeOIDS[i:i+5])
            break

if TotalRead != writeOIDS:
    print "*** TEST FAILED - Read OIDS != Write OIDS ***"

#print " OIDS Written to the Queue -------------------"
#print writeOIDS
#print " OIDS Read Back from the Queue ---------------"
#print readOIDS
